%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% rush_hour/solver.pl
%
% Rush Hour puzzle solver
% (C) 2019 Norbert Zeh (nzeh@cs.dal.ca)
%
% The solver logic
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


% Import the state predicates
:- [rush_hour/state].


% Solve the puzzle
solve_puzzle(Puzzle, Moves) :-
  puzzle_state(Puzzle, State),
  bfs(State, Moves).


% Run BFS from the given state to find a shortest solution path
bfs(State, Moves) :-
  bfs_level([State], [], _, Moves).


% A predicate that implements the level-by-level BFS search for a solution.
bfs_level(CurrentLevel, _, SolvedState, []) :-
  CurrentLevel = [_|_],
  findall(State, (member(State, CurrentLevel), state_is_solved(State)), [SolvedState|_]),
  !.

bfs_level(CurrentLevel, PreviousLevel, SuccessorState, Moves) :-
  CurrentLevel = [_|_],
  maplist(neighbours, CurrentLevel, Edges0),
  append(Edges0, Edges),
  maplist(nth1(2), Edges, Neighbours0),
  list_to_ord_set(Neighbours0, Neighbours),
  ord_union(CurrentLevel, PreviousLevel, VisitedStates),
  ord_subtract(Neighbours, VisitedStates, NextLevel),
  bfs_level(NextLevel, CurrentLevel, SuccSuccState, SuccMoves),
  member([SuccessorState, SuccSuccState, Move], Edges),
  Moves = [Move|SuccMoves], !.


% Find the edges from a state to all states that can be reached from it using a
% single move.
neighbours(State, Neighbours) :-
  findall([State, Neighbour, Move],
          ( between(0, 63, Pos),
            between(-4, 4, Offset), Offset \= 0,
            move(State, Pos, Offset, Neighbour, Move)
          ),
          Neighbours).


% Try to perform a move of the piece at the given position by the given offset
move(State, Pos, Offset, Neighbour, Move) :-
  move0(State, Pos, Offset, Neighbour),
  pos_offset_move(Pos, Offset, Move).


% Helper for constructing the new state resulting from a move.
move0(State, Pos, Offset, Neighbour) :-
  horizontal_move(State, Pos, Offset, Neighbour), !.

move0(State, Pos, Offset, Neighbour) :-
  vertical_move(State, Pos, Offset, Neighbour), !.
