;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; rush-hour/solver.ss
;;;
;;; Rush Hour puzzle solver
;;; (C) 2019 Norbert Zeh (nzeh@cs.dal.ca)
;;;
;;; Implementation of the search for a shortest move sequence that solves a
;;; given Rush Hour puzzle.
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(library (rush-hour solver)
  (export solve-puzzle)
  (import (rnrs (6))
          (rush-hour state)
          (rush-hour utils))


  ;; Solve a Rush Hour puzzle represented as a 36-character string listing the
  ;; 6x6 cells of the board in row-major order.  Empty cells are marked with "o".
  ;; Occupied cells are marked  with letters representing pieces.  Cells occupied
  ;; by the same piece carry the same letter.
  (define (solve-puzzle puzzle)
    (let ([start-state (state-from-string-rep puzzle)]
          [seen-states (make-hashtable equal-hash equal?)])

      (hashtable-set! seen-states start-state #t)

      (let bfs ([current-level `((,start-state . ()))])

        (let* ([neighbours (apply append (map extend-path current-level))]
               [next-level (filter (partial is-new? seen-states) neighbours)]
               [solution   (find is-solution? next-level)])

               (cond [solution           (reverse (cdr solution))]
                     [(null? next-level) #f]
                     [else               (bfs next-level)])))))


  ;; See whether the given path (sequence of moves) can be extended with a single
  ;; move to obtain a solution.  If so, return this solution.  Otherwise, return
  ;; the set of new states discovered.
  (define (extend-path path)
    (let ([extensions
           (apply append
                  (map-for pos from 0 to 63
                    (append (map-for offset from 1 to 4
                              (extend-path-with-move path pos offset))
                            (map-for offset from -1 downto -4
                              (extend-path-with-move path pos offset)))))])
      (filter id extensions)))


  ;; Try to apply the given move to the given state.  If this is a valid move,
  ;; the path is extended.  Otherwise, #f is returned.
  (define (extend-path-with-move path pos offset)
    (let* ([state     (car path)]
           [new-state (cond [(not (state-is-end? state pos))
                             #f]
                            [(state-is-horizontal? state pos)
                             (state-horizontal-move state pos offset)]
                            [else
                             (state-vertical-move state pos offset)])])
      (and new-state
           (cons new-state
                 (cons (state-make-move pos offset)
                       (cdr path))))))


  ;; Check whether the given path is a solution
  (define (is-solution? path)
    (state-is-solved? (car path)))


  ;; Check whether the given state is new
  (define (is-new? seen-states path)
    (and (not (hashtable-contains? seen-states (car path)))
         (begin (hashtable-set! seen-states (car path) #t)
                #t))))
